#include "scanconverter.h"
#include "gillib.h"

BOOL CScanConverter::m_bWindingEnabled = TRUE;

//////////////////////////////////////////////////////////////////////
// CScanConverter Class
//

CScanConverter::CScanConverter()
{
  m_uiWidth = 0;
  m_uiHeight = 0;

  m_pcolBuffer = NULL;
  m_pzBuffer = NULL;
}

CScanConverter::~CScanConverter()
{
  Uninit();
}

VOID CScanConverter::Init()
{
  m_uiWidth = 0;
  m_uiHeight = 0;

  m_pcolBuffer = NULL;
  m_pzBuffer = NULL;
}

VOID CScanConverter::Uninit()
{

  if ( m_pcolBuffer != NULL )
    {
      delete m_pcolBuffer;
      m_pcolBuffer = NULL;
    }

  if ( m_pzBuffer != NULL )
    {
      delete m_pzBuffer;
      m_pzBuffer = NULL;
    }

  m_uiWidth = 0;
  m_uiHeight = 0;
}

VOID CScanConverter::HandleReshape(UINT_32 uiWidth, UINT_32 uiHeight)
{
  ASSERT( uiWidth > 0 && uiHeight > 0 );

  if ( m_pcolBuffer != NULL )
    {
      delete m_pcolBuffer;
      m_pcolBuffer = NULL;
    }

  if ( m_pzBuffer != NULL )
    {
      delete m_pzBuffer;
      m_pzBuffer = NULL;
    }

  m_uiWidth = uiWidth;
  m_uiHeight = uiHeight;

  m_pcolBuffer = new PIXELCOLOR[m_uiWidth*m_uiHeight];
  ASSERT( m_pcolBuffer != NULL );
  m_pzBuffer = new PIXELZ[m_uiWidth*m_uiHeight];
  ASSERT( m_pzBuffer != NULL );
}

VOID CScanConverter::ClearBuffers()
{
  ASSERT( m_pcolBuffer != NULL && m_pzBuffer != NULL );

  memset(m_pcolBuffer, 0, m_uiWidth*m_uiHeight*sizeof(PIXELCOLOR));
  memset(m_pzBuffer, 0, m_uiWidth*m_uiHeight*sizeof(PIXELZ));
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 7: Scan conversion and visibility
//
// Input: Contour list of vertices in 4D (before divide by W) canonical 
//        half cube
// Effect: Fill in the correct pixels using the current inside/outside
//         test (Parity or Winding Number) and the Z Buffer visibility
//         algorithm.  Use additive colors for pixel with identical Z
//         values to help debug your boundary cases.
VOID CScanConverter::FillPolygon(const Matrix &mGvp_proj, const CListInt &listVertices)
{
  Vector vColor;
  PIXELCOLOR col, *pcolLine, *pcol;
  UINT_32 i, j, uiXMin, uiXMax, uiYMin, uiYMax;

  ASSERT( m_pcolBuffer != NULL && m_pzBuffer != NULL );

  // Inside / Outside Test
  if ( GetWindingEnabled() )
    {
      // Use Winding Number Rule 
    }
  else 
    {
      // Use the Parity Rule
    }

//////////////////////////////////////////////////
// START: Sample fill using additive colors
//
  // Red Rectangle
  vColor.Set(1.0, 0.0, 0.0);
  PackColor(col, vColor);

  uiXMin = m_uiWidth / 8;
  uiXMax = (5 * m_uiWidth) / 8;
  uiYMin = m_uiHeight / 8;
  uiYMax = (5 * m_uiHeight) / 8;

  pcolLine = &(m_pcolBuffer[uiYMin * m_uiWidth + uiXMin]);
  for ( i = uiYMin; i < uiYMax; i++ )
    {
      pcol = pcolLine;
      for ( j = uiXMin; j < uiXMax; j++ )
	{
	  pcol->r += col.r;
	  pcol->g += col.g;
	  pcol->b += col.b;
	  pcol++;
	}
      pcolLine += m_uiWidth;
    }

  // Green Rectangle
  vColor.Set(0.0, 1.0, 0.0);
  PackColor(col, vColor);

  uiXMin = (3 * m_uiWidth) / 8;
  uiXMax = (7 * m_uiWidth) / 8;
  uiYMin = (3 * m_uiHeight) / 8;
  uiYMax = (7 * m_uiHeight) / 8;

  pcolLine = &(m_pcolBuffer[uiYMin * m_uiWidth + uiXMin]);
  for ( i = uiYMin; i < uiYMax; i++ )
    {
      pcol = pcolLine;
      for ( j = uiXMin; j < uiXMax; j++ )
	{
	  pcol->r += col.r;
	  pcol->g += col.g;
	  pcol->b += col.b;
	  pcol++;
	}
      pcolLine += m_uiWidth;
    }
//
// END  : Sample fill using additive colors
//////////////////////////////////////////////////
}
//
// START: ASSIGNMENT 7: Scan conversion and visibility
//////////////////////////////////////////////////////////////////////////////

VOID CScanConverter::RenderPixels()
{
  ASSERT( m_pcolBuffer != NULL && m_pzBuffer != NULL );

  gilDrawPixels(m_uiWidth, m_uiHeight, GL_RGB, GL_UNSIGNED_BYTE,
		m_pcolBuffer);
}



