#ifndef __SCANCONVERTER_H__
#define __SCANCONVERTER_H__

#include "mathlib.h"
#include "datalib.h"

typedef struct _PIXELCOLOR
{
  UINT_8 r;
  UINT_8 g;
  UINT_8 b;
} PIXELCOLOR;

typedef UINT_8 PIXELZ;

//////////////////////////////////////////////////////////////////////
// CScanConverter Class
//

class CScanConverter
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CScanConverter();
  ~CScanConverter();

  VOID Init();
  VOID Uninit();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  // Called by the Window when its dimensions change
  VOID HandleReshape(UINT_32 uiWidth, UINT_32 uiHeight);

//////////////////////////////////////////////////
// Rendering
//
public:
  // Clears out the pixel and Z buffers to all zeroes
  VOID ClearBuffers();
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 7: Scan conversion and visibility
//
  // Input: Contour list of vertices in 4D (before divide by W) canonical 
  //        half cube
  // Effect: Fill in the correct pixels using the current inside/outside
  //         test (Parity or Winding Number) and the Z Buffer visibility
  //         algorithm.  Use additive colors for pixel with identical Z
  //         values to help debug your boundary cases.
  VOID FillPolygon(const Matrix &mGvp_proj, const CListInt &listVertices);
//
// END  : ASSIGNMENT 7: Scan conversion and visibility
//////////////////////////////////////////////////////////////////////////////
  // Transfers the pixel buffer to OpenGL (like glSwapBuffers)
  VOID RenderPixels();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  static inline BOOL GetWindingEnabled();
  static inline VOID SetWindingEnabled(BOOL bWindingEnabled);

//////////////////////////////////////////////////
// Class Variables
//
public:
  static BOOL m_bWindingEnabled; // Use the Winding Number Rule?

//////////////////////////////////////////////////
// Member Variables
//
protected:
  UINT_32 m_uiWidth;  // The width of the window
  UINT_32 m_uiHeight; // The height of the window

  PIXELCOLOR *m_pcolBuffer; // The frame buffer (Color Pixels)
  PIXELZ *m_pzBuffer;       // The Z buffer
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline BOOL CScanConverter::GetWindingEnabled()
{
  return m_bWindingEnabled;
}

inline VOID CScanConverter::SetWindingEnabled(BOOL bWindingEnabled)
{
  m_bWindingEnabled = bWindingEnabled;
}

inline VOID PackColor(PIXELCOLOR &col, const Vector &vColor)
{
  col.r = (UINT_8)(((1<<8)-1)*vColor(0));
  col.g = (UINT_8)(((1<<8)-1)*vColor(1));
  col.b = (UINT_8)(((1<<8)-1)*vColor(2));
} 

inline VOID PackZ(PIXELZ &z, FLOAT fZ)
{
  z = (UINT_8)(((1<<8)-1)*(1.0 + fZ));
} 

#endif // __SCANCONVERTER_H__


