#ifndef __SLIDEGEOMETRY_H__
#define __SLIDEGEOMETRY_H__

#include "genlib.h"
#include "datalib.h"

#include "slide.h"
#include "gillib.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEGeometry Class
//

class CSLIDEGeometry
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEGeometry();
  virtual ~CSLIDEGeometry();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render() = 0;

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile) = 0;

//////////////////////////////////////////////////
// Error Handling
//
public:
  VOID HandleError(CHAR *pcFormat, ...);
  VOID HandleWarning(CHAR *pcFormat, ...);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  static VOID SetErrorFile(FILE *fpErrorFile);
  static UINT_32 GetErrorCount();
  static VOID SetErrorCount(UINT_32 uiErrors);
  static UINT_32 GetWarningCount();
  static VOID SetWarningCount(UINT_32 uiWarnings);

public:
  inline CHAR *GetFileName();
  inline UINT_32 GetFileLine();
  inline UINT_32 GetFileColumn();
  inline CHAR *GetID();

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static FILE *m_fpErrorFile;
  static UINT_32 m_uiErrorCount;
  static UINT_32 m_uiWarningCount;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // SLIDE .slf file position of the geometry for debugging the files
  CHAR *m_pcFile;
  UINT_32 m_uiLine;
  UINT_32 m_uiColumn;

  CHAR *m_pcID; // Stored in a string table so equals checks can be done on 
		// the pointer itself
};

//////////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR *CSLIDEGeometry::GetFileName()
{
  return m_pcFile;
}

inline UINT_32 CSLIDEGeometry::GetFileLine()
{
  return m_uiLine;
}

inline UINT_32 CSLIDEGeometry::GetFileColumn()
{
  return m_uiColumn;
}

inline CHAR *CSLIDEGeometry::GetID()
{
  return m_pcID;
}

//////////////////////////////////////////////////////////////////////////////
// CSLIDEGeometryIter Class
//

class CSLIDEGeometryIter
{
//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEGeometryIter();
  virtual ~CSLIDEGeometryIter();
  virtual VOID Init(CSLIDEGeometry *pCSLIDEGeometry);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CSLIDEGeometry *m_pCSLIDEGeometry;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif __SLIDEGEOMETRY_H__
