#include "slideprimitive.h"

FILE *CSLIDEPrimitive::m_fpErrorFile = stderr;
UINT_32 CSLIDEPrimitive::m_uiErrorCount = 0;
UINT_32 CSLIDEPrimitive::m_uiWarningCount = 0;

CSLIDEPrimitive::CSLIDEPrimitive()
{
  m_pcFile = NULL;
  m_uiLine = 0;
  m_uiColumn = 0;

  m_pcID = NULL;
  m_bDynamic = FALSE;
}

CSLIDEPrimitive::~CSLIDEPrimitive()
{
  Uninit();
}

VOID CSLIDEPrimitive::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  ASSERT( pcFile != NULL );

  m_pcFile = pcFile;
  m_uiLine = uiLine;
  m_uiColumn = uiColumn;

  m_pcID = pcID;
}

VOID CSLIDEPrimitive::Uninit()
{
  m_pcID = NULL;
}

BOOL CSLIDEPrimitive::Initialized()
{
  return m_pcFile != NULL;
}

BOOL CSLIDEPrimitive::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  ASSERT( pCSLIDEWorld != NULL );

  return TRUE;
}

BOOL CSLIDEPrimitive::Valid()
{
  return Initialized();
}

VOID CSLIDEPrimitive::Preprocess()
{
  CheckDynamic();
  UpdateState();
}

VOID CSLIDEPrimitive::Update()
{
  if ( GetDynamic() )
    {
      UpdateState();
    }
}

VOID CSLIDEPrimitive::CheckDynamic()
{
  SetDynamic(FALSE);
}

VOID CSLIDEPrimitive::UpdateState()
{
}

VOID CSLIDEPrimitive::HandleError(CHAR *pcFormat, ...)
{
  va_list vaArgs;

  ASSERT( pcFormat != NULL );

  m_uiErrorCount++;
  fprintf(m_fpErrorFile, "%s: %ld: %ld: Semantic Error: ", 
	  (m_pcFile == NULL) ? ("") : (m_pcFile), m_uiLine, m_uiColumn);

  // Now print the message sent by the program
  va_start(vaArgs, pcFormat);
  vfprintf(m_fpErrorFile, pcFormat, vaArgs);
  va_end(vaArgs);

  fprintf(m_fpErrorFile, "\n"); 
}

VOID CSLIDEPrimitive::HandleWarning(CHAR *pcFormat, ...)
{
  va_list vaArgs;

  ASSERT( pcFormat != NULL );

  m_uiWarningCount++;
  fprintf(m_fpErrorFile, "%s: %ld: %ld: Semantic Warning: ", 
	  (m_pcFile == NULL) ? ("") : (m_pcFile), m_uiLine, m_uiColumn);

  // Now print the message sent by the program
  va_start(vaArgs, pcFormat);
  vfprintf(m_fpErrorFile, pcFormat, vaArgs);
  va_end(vaArgs);

  fprintf(m_fpErrorFile, "\n"); 
}

//////////////////////////////////////////////////////////////////////
// CSLIDEPrimitiveIter Class
//

CSLIDEPrimitiveIter::CSLIDEPrimitiveIter()
{
  m_pCSLIDEPrimitive = NULL;
}

CSLIDEPrimitiveIter::~CSLIDEPrimitiveIter()
{
  Uninit();
}

VOID CSLIDEPrimitiveIter::Init(CSLIDEPrimitive *pCSLIDEPrimitive)
{
  ASSERT( pCSLIDEPrimitive != NULL );
  ASSERT( pCSLIDEPrimitive->Initialized() );

  m_pCSLIDEPrimitive = pCSLIDEPrimitive;
}

VOID CSLIDEPrimitiveIter::Uninit()
{
  m_pCSLIDEPrimitive = NULL;
}

BOOL CSLIDEPrimitiveIter::Initialized()
{
  return (m_pCSLIDEPrimitive != NULL);
}

BOOL CSLIDEPrimitiveIter::Valid()
{
  if ( !Initialized() )
    {
      return FALSE;
    }

  return m_pCSLIDEPrimitive->Initialized();
}





