#include "sliderender.h"

#include "slideworld.h"
#include "slideviewport.h"
#include "slidecamera.h"
#include "slidenode.h"
#include "slideface.h"
#include "slidesurface.h"

static CHAR gpcBufferVariable[200];
static CHAR gpcBufferValue[200];

BOOL CSLIDERender::m_bRenderClipBound = FALSE;
BOOL CSLIDERender::m_bCrystalBallEnabled = FALSE;
BOOL CSLIDERender::m_bRotateClippedEnabled = FALSE;

UINT_32 CSLIDERender::m_uiCount = 0;
Vector CSLIDERender::m_vProjMin;
Vector CSLIDERender::m_vProjMax;
CBound CSLIDERender::m_boundClip;

CListInt *CSLIDERender::m_plistCSLIDETclVariables = NULL;

//////////////////////////////////////////////////////////////////////
// CSLIDERender Class
//

CSLIDERender::CSLIDERender()
{
  m_pcTclTag = NULL;
  m_pCSLIDETclVariables = NULL;

  m_pcViewportID = NULL;

  m_pCSLIDECameraPath = NULL;

  m_pcWorldRootID = NULL;
  m_pWorldRoot = NULL;

  m_dyniLOD.SetValue(slidelodtSLF_FULL);

  if ( m_uiCount == 0 )
    {
      // These are the limits of the canonical view volume
      m_vProjMin.Set(-1, -1, -1);
      m_vProjMax.Set(1, 1, 0);

      // This a bound around the canonical view volume which
      // we will use to do clipping with homogeneous coordinates
      Point pt;
      
      pt = m_vProjMin;
      m_boundClip.Insert(pt);
      pt = m_vProjMax;
      m_boundClip.Insert(pt);

      m_plistCSLIDETclVariables = new CListInt;
      ASSERT( m_plistCSLIDETclVariables != NULL );
      m_plistCSLIDETclVariables->Init();
    }
  m_uiCount++;
}

CSLIDERender::~CSLIDERender()
{
  CSLIDETclVariables *pCSLIDETclVariables;

  m_uiCount--;
  if ( m_uiCount == 0 )
    {
      if ( m_plistCSLIDETclVariables == NULL )
	{
	  for ( pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->RemoveFirst();
		pCSLIDETclVariables != NULL;
		pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->RemoveFirst() )
	    {
	      delete pCSLIDETclVariables;
	    }
	  
	  delete m_plistCSLIDETclVariables;
	  m_plistCSLIDETclVariables = NULL;
	}
    }
}

VOID CSLIDERender::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcViewportID, CSLIDECameraPath *pCSLIDECameraPath,
			CHAR *pcWorldRootID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, NULL);

  ASSERT( pcViewportID != NULL );
  ASSERT( pCSLIDECameraPath != NULL );
  ASSERT( pcWorldRootID != NULL );

  m_pcTclTag = NULL;
  m_pCSLIDETclVariables = NULL;

  m_pcViewportID = pcViewportID;

  m_pCSLIDECameraPath = pCSLIDECameraPath;

  m_pcWorldRootID = pcWorldRootID;
  m_pWorldRoot = NULL;

  m_dyniLOD.SetValue(slidelodtSLF_FULL);

  // Crystal Ball Variables
  m_iButtonCurr = GLSI_INVALID_BUTTON;
  m_fScaleLast = 1.0;
}

VOID CSLIDERender::Uninit()
{
  CSLIDEPrimitive::Uninit();

  m_pcViewportID = NULL;

  if ( m_pCSLIDECameraPath != NULL )
    {
      delete m_pCSLIDECameraPath;
      m_pCSLIDECameraPath = NULL;
    }

  m_pcWorldRootID = NULL;
  m_pWorldRoot = NULL;
}

BOOL CSLIDERender::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcViewportID != NULL &&
	   m_pCSLIDECameraPath != NULL &&
           m_pcWorldRootID != NULL );
}

BOOL CSLIDERender::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect, b;
  CSLIDEViewport *pCSLIDEViewport;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Tcl Tag to see if we are sharing UI Variables

  m_pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->Peek((INT_32)m_pcTclTag);
  if ( m_pCSLIDETclVariables == NULL )
    {
      m_pCSLIDETclVariables = new CSLIDETclVariables;
      ASSERT( m_pCSLIDETclVariables != NULL);
      m_pCSLIDETclVariables->Init(m_pcTclTag);

      m_plistCSLIDETclVariables->Insert(m_pCSLIDETclVariables, 
					(INT_32)m_pcTclTag);
    }

  // Check Viewport Reference

  pCSLIDEViewport = pCSLIDEWorld->PeekViewport(m_pcViewportID);
  if ( pCSLIDEViewport == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtViewport],
		  m_pcViewportID);
      bCorrect = FALSE;
    }
  else
    {
      pCSLIDEViewport->InsertRender(this);
    }

  // Check Camera Reference

  b = m_pCSLIDECameraPath->CheckSemantics(pCSLIDEWorld);
  bCorrect = bCorrect && b;

  // Check Root Node Reference

  m_pWorldRoot = pCSLIDEWorld->PeekNode(m_pcWorldRootID);
  if ( m_pWorldRoot == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  m_pcWorldRootID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDERender::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return (m_pWorldRoot != NULL &&
	  ( 0 <= m_dyniLOD.GetValue() &&
	    m_dyniLOD.GetValue() < slidelodtSize ) );
}

VOID CSLIDERender::PreprocessGraph()
{
  m_pWorldRoot->PreprocessGraph();
}

VOID CSLIDERender::UpdateGraph()
{
  m_pWorldRoot->UpdateGraph();
}

VOID CSLIDERender::CheckDynamic()
{
  BOOL bDynamic;

  bDynamic = m_dyniLOD.GetDynamic();

  m_pCSLIDECameraPath->Preprocess();
  bDynamic = bDynamic || m_pCSLIDECameraPath->GetDynamic();

  SetDynamic(bDynamic);
}

VOID CSLIDERender::UpdateState()
{
  m_dyniLOD.Update();
  if ( !( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slidelodtSize ) )
    {
      HandleError("Illegal LOD flag value:  %ld", 
		  m_dyniLOD.GetValue());
      m_dyniLOD.SetValue(slidelodtSLF_FULL);
    }

  if ( m_dyniLOD.GetValue() != slidelodtSLF_OFF )
    {
      m_pCSLIDECameraPath->Update();
    }
}

VOID CSLIDERender::HandleReshape(const Vector &vSize)
{
  if ( m_dyniLOD.GetValue() != slideoftSLF_OFF )
    {
      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
    }
}

// Both Mouse Functions get the size of the viewport rectangle
// and the location of the mouse cursor relative to the lower 
// left hand corner
BOOL CSLIDERender::HandleMouseButton(const Vector &vSize,
				     INT_32 iButton, INT_32 iState, 
				     const Vector &vClick)
{
  BOOL bPostRedisplay;

  bPostRedisplay = FALSE;

  if ( iState == GLSI_DOWN )
    {
      if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
	{
	  return FALSE;
	}

      m_vStart = vClick - (vSize / 2.0);
      m_vStart *= (FLOAT)(2.0 / MIN(vSize(X), vSize(Y)));
      m_vLatest = m_vStart;
      m_iButtonCurr = iButton;

      switch ( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  break;
	case GLSI_MIDDLE_BUTTON:
	  m_fScaleLast = 1.0;
	  break;
	case GLSI_RIGHT_BUTTON:
	  break;
	default:
	  ASSERT(FALSE);
	  break;
	}
    }
  else if ( iState == GLSI_UP && iButton == m_iButtonCurr )
    {
      m_iButtonCurr = GLSI_INVALID_BUTTON;
    }

  return bPostRedisplay;
}

// We change the vClick into a vector from the center of the window
// where the radius of the best fit sphere is 1.0
BOOL CSLIDERender::HandleMouseMotion(const Vector &vSize,
				     const Vector &vClick)
{
  BOOL bPostRedisplay;

  bPostRedisplay = FALSE;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      m_iButtonCurr = GLSI_INVALID_BUTTON;
      return FALSE;
    }

  if ( m_iButtonCurr != GLSI_INVALID_BUTTON )
    {
      Matrix m, mInverse;
      Vector vCurr, v1, v2, vNorm;
      Quater q;
      FLOAT theta;

      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
      m_pCSLIDETclVariables->SetMouseX(vClick(X)/vSize(X));
      m_pCSLIDETclVariables->SetMouseY(vClick(Y)/vSize(Y));
      
      vCurr = vClick - (vSize / 2.0);
      vCurr *= (FLOAT)(2.0 / MIN(vSize(X), vSize(Y)));

      switch ( m_iButtonCurr ) 
	{
	case GLSI_LEFT_BUTTON:
	  {
	    // General Rotations
	    
	    v1.Set(m_vLatest[X], m_vLatest[Y], 
		   (FLOAT)(1.0 - (m_vLatest[X]*m_vLatest[X] + m_vLatest[Y]*m_vLatest[Y])));
	    v2.Set(vCurr[X], vCurr[Y], 
		   (FLOAT)(1.0 - (vCurr[X]*vCurr[X] + vCurr[Y]*vCurr[Y])));
	    
	    if ( v1[Z] >= 0.0f && v2[Z] >= 0.0f )
	      {
		v1[Z] = (FLOAT)sqrt(v1[Z]);
		v2[Z] = (FLOAT)sqrt(v2[Z]);
	      }
	    else
	      {
		v1[Z] = 0.0;
		v2[Z] = 0.0;
	      }
	    
	    VERIFY(v1.Normalize());
	    VERIFY(v2.Normalize());
	    
	    theta = (FLOAT)MAX(MIN(v1.Dot(v2), 1.0), -1.0);
	    theta = (FLOAT)acos(theta);
	    ASSERT( theta >= 0.0f );

	    vNorm = v1 * v2;
	    
	    if ( vNorm.Normalize() )
	      {
		m.SetRotate(vNorm, theta);
		mInverse.SetRotate(vNorm, -theta);

		bPostRedisplay = TRUE;
	      }

	    q.SetRotation(v1, v2);	    
	    m_pCSLIDETclVariables->SetRotate(
				     q*m_pCSLIDETclVariables->GetRotate()); 
	  }
	  break;
	case GLSI_MIDDLE_BUTTON:
	  {
	    FLOAT fScale;

	    // Scaling
	    ASSERT( m_fScaleLast > 0.0 );
	    fScale = (FLOAT)pow(2, 5*(vCurr[Y] - m_vStart[Y])/2.0);
	    m.SetScale(fScale / m_fScaleLast);
	    mInverse.SetScale(m_fScaleLast / fScale);

	    m_pCSLIDETclVariables->SetScale(m_pCSLIDETclVariables->GetScale() * 
					    fScale / m_fScaleLast);

	    bPostRedisplay = TRUE;

	    m_fScaleLast = fScale;
	  }
	  break;
	case GLSI_RIGHT_BUTTON:
	  {
	    // Rotations around the Z-axis

	    v1 = m_vLatest;
	    v2 = vCurr;
	    
	    v1.Normalize();
	    v2.Normalize();

	    theta = (FLOAT)MAX(MIN(v1.Dot(v2), 1.0), -1.0);
	    theta = (FLOAT)acos(theta);
	    ASSERT( theta >= 0.0f );
	    
	    vNorm = v1 * v2;
	    if ( vNorm.Normalize() )
	      {
		m.SetRotate(vNorm, theta);
		mInverse.SetRotate(vNorm, -theta);
	      }
	    else
	      {		    
		m.SetRotateZ(theta);
		    mInverse.SetRotateZ(-theta);
	      }
	    
	    q.SetRotation(v1, v2);	    
	    m_pCSLIDETclVariables->SetRotate(
				     q*m_pCSLIDETclVariables->GetRotate()); 

	    if ( theta != 0.0 )
	      {
		bPostRedisplay = TRUE;
	      }
	  }
	  break;
	}

      // Update the Matrix if it has changed
      if ( bPostRedisplay )
	{
	  Matrix mGvrp_vrc, mGvrc_vrp;

	  mGvrp_vrc = m_pCSLIDECameraPath->GetCamera()->GetVRCToVRP();
	  mGvrc_vrp = m_pCSLIDECameraPath->GetCamera()->GetVRPToVRC();

	  // Bring the Center of the VRC to the origin before applying scales and 
	  // rotates
	  m_mBall = m_pCSLIDECameraPath->GetGroot_obj() * mGvrc_vrp * m * 
	            mGvrp_vrc * m_pCSLIDECameraPath->GetGobj_root() * m_mBall;
	  m_mBallInverse = m_mBallInverse * 
	                   m_pCSLIDECameraPath->GetGroot_obj() * mGvrc_vrp * 
	                   mInverse * 
	                   mGvrp_vrc * m_pCSLIDECameraPath->GetGobj_root();

	  // Even if the world is paused we want the UI to take effect
	  CSLIDEWorld::GetWorld()->ScheduleDynamics(0);
	}

      m_vLatest = vCurr;
    }

  return bPostRedisplay;
}

VOID CSLIDERender::Render(const Vector &vVPOrigin, const Vector &vVPSize)
{
  Matrix mGvp_proj, mGvrc_obj, mGobj_vrc;
  Vector vFrustum, vVPMax;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      return;
    }

  // Set the NormalRenderLength based on the camera's Frustum
  vFrustum = m_pCSLIDECameraPath->GetCamera()->GetFrustumMax() -
             m_pCSLIDECameraPath->GetCamera()->GetFrustumMin();
  CSLIDEFace::SetNormalsRenderLength((FLOAT)(0.05 * MIN(vFrustum[X], vFrustum[Y])));


  // Get the Viewport Boundary
  vVPMax = vVPOrigin + vVPSize;

  // Create the matrix that goes from the canonical view volume
  // to the Viewport
  mGvp_proj.SetViewPort(vVPOrigin, vVPMax,
			m_vProjMin, m_vProjMax, 
			vFrustum[Y]/vFrustum[X]);

  if ( GetRotateClippedEnabled() )
    {
      Vector vTrans;
      Matrix mTrans, mTransInv;

      vTrans.Set(0.0, 0.0, 0.5);
      mTrans.SetTranslate(vTrans);

      mTransInv.SetTranslate(-vTrans);

      mGvp_proj = mGvp_proj * mTransInv * m_mBall * mTrans;
    }

  // The mGproj_vrc is the Camera's projection matrix

  // These lines look backward but they are not
  // we need the inverse of the camera path
  mGvrc_obj = m_pCSLIDECameraPath->GetGobj_root();
  mGobj_vrc = m_pCSLIDECameraPath->GetGroot_obj();

  if ( GetCrystalBallEnabled() && !GetRotateClippedEnabled() )
    {
      mGvrc_obj = mGvrc_obj * m_mBall;
      mGobj_vrc = m_mBallInverse * mGobj_vrc;
    }

  //CSLIDESurface::GetSLF_INHERIT()->Render();
  m_pWorldRoot->Render(mGvp_proj, 
		       m_pCSLIDECameraPath->GetCamera()->GetVRCToProjection(), 
		       m_pCSLIDECameraPath->GetCamera()->GetProjectionToVRC(), 
		       mGvrc_obj, mGobj_vrc,
		       &m_boundClip, 
		       NULL,
		       CSLIDESurface::GetSLF_INHERIT(),
		       slideshtSLF_INHERIT, 
		       (SLIDELODType)m_dyniLOD.GetValue());

  // Render Clip Bound
  if ( GetRenderClipBound() )
    {
      gilColor3f(1.0, 1.0, 1.0);
      m_boundClip.Render(mGvp_proj);
    }
}

VOID CSLIDERender::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ", 
	  gcppcSLIDEReservedWords[sliderwtRender], 
	  m_pcViewportID);
  m_pCSLIDECameraPath->WriteSLIDE(fpOutputFile);
  fprintf(fpOutputFile, " %s\n", m_pcWorldRootID);

  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndRender]);
}

//////////////////////////////////////////////////////////////////////
// CSLIDETclVariables Class
//

CSLIDETclVariables::CSLIDETclVariables()
{
  m_pcID = NULL;

  m_fWidth = 1.0;
  m_fHeight = 1.0;
  m_fMouseX = 0.0;
  m_fMouseY = 0.0;
  m_fScale = 1.0;
}

CSLIDETclVariables::~CSLIDETclVariables()
{
  Uninit();
}

VOID CSLIDETclVariables::Init(CHAR *pcID)
{
  Vector v;
  Quater q;

  m_pcID = pcID;

  SetWidth(1.0);
  SetHeight(1.0);
  SetMouseX(0.0);
  SetMouseY(0.0);

  v.Set(0.0, 0.0, 1.0);
  q.SetRotation(v, 0.0);

  SetRotate(q);
  SetScale(1.0);
}

VOID CSLIDETclVariables::Uninit()
{
}

BOOL CSLIDETclVariables::Initialized()
{
  return TRUE;
}

BOOL CSLIDETclVariables::Valid()
{
  return TRUE;
}

VOID CSLIDETclVariables::SetWidth(FLOAT fWidth)
{
  m_fWidth = fWidth;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_WIDTH");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_WIDTH", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fWidth);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetHeight(FLOAT fHeight)
{
  m_fHeight = fHeight;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_HEIGHT");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_HEIGHT", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fHeight);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetMouseX(FLOAT fMouseX)
{
  m_fMouseX = fMouseX;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_MOUSE_X");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_MOUSE_X", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fMouseX);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetMouseY(FLOAT fMouseY)
{
  m_fMouseY = fMouseY;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_MOUSE_Y");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_MOUSE_Y", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fMouseY);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetRotate(const Quater &qRotate)
{
  Vector v;
  FLOAT fTheta;

  m_qRotate = qRotate;

  m_qRotate.GetRotation(v, fTheta);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_AXIS_X");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_AXIS_X", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", v[X]);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_AXIS_Y");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_AXIS_Y", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", v[Y]);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_AXIS_Z");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_AXIS_Z", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", v[Z]);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_ANGLE");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_ANGLE", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", RADTODEG(fTheta));
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetScale(FLOAT fScale)
{
  m_fScale = fScale;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_SCALE");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_SCALE", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fScale);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

