#ifndef __SLIDEVIEWPORT_H__
#define __SLIDEVIEWPORT_H__

#include "slideprimitive.h"

class CSLIDEWorld;
class CSLIDERender;

//////////////////////////////////////////////////////////////////////
// CSLIDEViewport Class
//

class CSLIDEViewportIter;

class CSLIDEViewport : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend class CSLIDEViewportIter;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEViewport();
  virtual ~CSLIDEViewport();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, CHAR *pcWindowID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

public:  
  virtual VOID InsertRender(CSLIDERender *pCSLIDERender);

//////////////////////////////////////////////////
// Iteration Functions
//
public:
  virtual CSLIDEPrimitiveIter *AllocateIter();
  virtual VOID FreeIter(CSLIDEPrimitiveIter *pCSLIDEViewportIter);

//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:
  VOID CheckDynamic();
  VOID UpdateState();

//////////////////////////////////////////////////
// Input Functions
//
public:
  VOID HandleReshape(UINT_32 uiWidth, UINT_32 uiHeight);
  BOOL HandleMouseButton(UINT_32 uiWidth, UINT_32 uiHeight,
			 INT_32 iButton, INT_32 iState, 
			 INT_32 iX, INT_32 iY);
  BOOL HandleMouseMotion(UINT_32 uiWidth, UINT_32 uiHeight,
			 INT_32 iX, INT_32 iY);

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render(INT_32 iOriginX, INT_32 iOriginY,
		      UINT_32 uiWidth, UINT_32 uiHeight);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual UINT_32 GetRenderCount();
  inline VOID SetOrigin(FLOAT fX, FLOAT fY);
  inline VOID SetOrigin(DynVector *pdynv);
  inline VOID SetSize(FLOAT fX, FLOAT fY);
  inline VOID SetSize(DynVector *pdynv);

  static inline BOOL GetRenderViewportBound();
  static inline VOID SetRenderViewportBound(BOOL bRenderViewportBound);

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static BOOL m_bRenderViewportBound;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcWindowID;

  DynVector m_vOrigin;
  DynVector m_vSize;
  CListInt m_listCSLIDERender;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEViewport::SetOrigin(FLOAT fX, FLOAT fY)
{
  m_vOrigin.Set(fX, fY, 0.0);
}

inline VOID CSLIDEViewport::SetOrigin(DynVector *pdynv)
{
  m_vOrigin.SetAll(pdynv);
}

inline VOID CSLIDEViewport::SetSize(FLOAT fX, FLOAT fY)
{
  m_vSize.Set(fX, fY, 0.0);
}

inline VOID CSLIDEViewport::SetSize(DynVector *pdynv)
{
  m_vSize.SetAll(pdynv);
}

inline BOOL CSLIDEViewport::GetRenderViewportBound()
{
  return m_bRenderViewportBound;
}

inline VOID CSLIDEViewport::SetRenderViewportBound(BOOL bRenderViewportBound)
{
  m_bRenderViewportBound = bRenderViewportBound;
}

//////////////////////////////////////////////////////////////////////
// CSLIDEViewportIter Class
//

class CSLIDEViewportIter : public CSLIDEPrimitiveIter
{
//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEViewportIter();
  virtual ~CSLIDEViewportIter();
  virtual VOID Init(CSLIDEViewport *pCSLIDEViewport);
  virtual VOID Uninit();

//////////////////////////////////////////////////
// Iteration Functions
//
public:
  virtual CSLIDERender *PeekFirstCSLIDERender();
  virtual CSLIDERender *PeekNextCSLIDERender();

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListIterInt m_iter;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//


#endif // __SLIDEVIEWPORT_H__




