#include "slideworld.h"

#include "sliderender.h"
#include "slideviewport.h"
#include "slidenode.h"
#include "slideinstance.h"
#include "slidetransform.h"
#include "slideface.h"
#include "slidepoint.h"
#include "slidesurface.h"

static CHAR gpcBuffer[200];
const FLOAT gcfFrameRate = 30.0;
const FLOAT gcfFrameTime = (FLOAT)1.0 / gcfFrameRate;

Tcl_Interp *CSLIDEWorld::m_pInterp = NULL;
CSLIDEWorld *CSLIDEWorld::m_pCSLIDEWorld = NULL;

DOUBLE CSLIDEWorld::m_dSecondsPerFrameRequested = gcfFrameTime;
DOUBLE CSLIDEWorld::m_dUpdateTime = 0;


//////////////////////////////////////////////////////////////////////
// CSLIDEWorld Class
//

CSLIDEWorld::CSLIDEWorld()
{
  CSLIDESurface::m_pCSLIDESurfaceDefault = NULL;
}

CSLIDEWorld::~CSLIDEWorld()
{
  Uninit();
}

VOID CSLIDEWorld::Init()
{
  m_listCSLIDERender.Init();
  m_listCWindow.Init();
  m_hashCWindow.Init(1);
  m_listCSLIDEViewport.Init();
  m_hashCSLIDEViewport.Init(3);
  m_listCSLIDENode.Init();
  m_hashCSLIDENode.Init(101);
  m_listCSLIDEInstance.Init();
  m_listCSLIDETransform.Init();
  m_listCSLIDEFace.Init();
  m_hashCSLIDEFace.Init(101);
  m_listCSLIDEPoint.Init();
  m_hashCSLIDEPoint.Init(101);
  m_listCSLIDESurface.Init();
  m_hashCSLIDESurface.Init(101);
  m_listCSLIDEGlobalInitTclBlock.Init();
  m_listCSLIDEGlobalUpdateTclBlock.Init();

  m_listDynamics.Init();

  InitDefaults();

  // Dynamics
  m_bDynamicsPaused = FALSE;
  m_dSecondsPerFrameRequested = gcfFrameTime;
  m_ttkWakeUpIDDynamics = 0;
  m_dTimeBase = 0.0;
  m_uiFrame = 0;
}

VOID CSLIDEWorld::InitDefaults()
{
  // Create Default Objects

  // Default Surface
  CSLIDESurface::m_pCSLIDESurfaceDefault = new CSLIDESurface;
  ASSERT( CSLIDESurface::m_pCSLIDESurfaceDefault != NULL );
  CSLIDESurface::m_pCSLIDESurfaceDefault->Init("", 0, 0, 
	(CHAR*)(gcppcSLIDEReservedWords[sliderwtSLF_INHERIT]));
  VERIFY(InsertSurface(CSLIDESurface::m_pCSLIDESurfaceDefault));  
}

VOID CSLIDEWorld::Uninit()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDESurface *pCSLIDESurface;

  UnscheduleDynamics();

  // Uninit Defaults
  if ( CSLIDESurface::m_pCSLIDESurfaceDefault != NULL )
    {
      // This is in the list of surfaces, so it will be deleted 
      // at the end of this function
      //delete CSLIDESurface::m_pCSLIDESurfaceDefault;
      CSLIDESurface::m_pCSLIDESurfaceDefault = NULL;
    }

  
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      delete pCSLIDERender;
    }

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      delete pCWindow;
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      delete pCSLIDEViewport;
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      delete pCSLIDENode;
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      delete pCSLIDEInstance;
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      delete pCSLIDETransform;
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      delete pCSLIDEFace;
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      delete pCSLIDEPoint;
    }

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      delete pCSLIDESurface;
    }
}

VOID CSLIDEWorld::UninitTcl()
{
  CListIterInt iter;
  CWindow *pCWindow;

  UnscheduleDynamics();

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      Tk_DoWhenIdle(CWindow::CBHandleDestroy, (ClientData)pCWindow);
    }
}

BOOL CSLIDEWorld::Initialized()
{
  return TRUE;
}

BOOL CSLIDEWorld::InsertRender(CSLIDERender *pCSLIDERender)
{
  ASSERT( pCSLIDERender != NULL );

  m_listCSLIDERender.InsertLast(pCSLIDERender, 0);

  return TRUE;
}

BOOL CSLIDEWorld::InsertWindow(CWindow *pCWindow)
{
  CHAR *pcID;

  ASSERT( pCWindow != NULL );

  pcID = pCWindow->GetSLIDEID();

  m_hashCWindow.Insert(pCWindow, (UINT_32)pcID);
  m_listCWindow.InsertLast(pCWindow, 0);

  return TRUE;
}

CWindow *CSLIDEWorld::PeekWindow(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CWindow*)(m_hashCWindow.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertViewport(CSLIDEViewport *pCSLIDEViewport)
{
  CHAR *pcID;

  ASSERT( pCSLIDEViewport != NULL );

  pcID = pCSLIDEViewport->GetID();

  m_hashCSLIDEViewport.Insert(pCSLIDEViewport, (UINT_32)pcID);
  m_listCSLIDEViewport.InsertLast(pCSLIDEViewport, 0);

  return TRUE;
}

CSLIDEViewport *CSLIDEWorld::PeekViewport(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEViewport*)(m_hashCSLIDEViewport.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertNode(CSLIDENode *pCSLIDENode)
{
  CHAR *pcID;

  ASSERT( pCSLIDENode != NULL );

  pcID = pCSLIDENode->GetID();

  m_hashCSLIDENode.Insert(pCSLIDENode, (UINT_32)pcID);
  m_listCSLIDENode.InsertLast(pCSLIDENode, 0);

  return TRUE;
}

CSLIDENode *CSLIDEWorld::PeekNode(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDENode*)(m_hashCSLIDENode.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.InsertLast(pCSLIDEInstance, 0);

  return TRUE;
}

BOOL CSLIDEWorld::InsertTransform(CSLIDETransform *pCSLIDETransform)
{
  ASSERT( pCSLIDETransform != NULL );

  m_listCSLIDETransform.InsertLast(pCSLIDETransform, 0);

  return TRUE;
}

BOOL CSLIDEWorld::InsertFace(CSLIDEFace *pCSLIDEFace)
{
  CHAR *pcID;

  ASSERT( pCSLIDEFace != NULL );

  pcID = pCSLIDEFace->GetID();

  m_hashCSLIDEFace.Insert(pCSLIDEFace, (UINT_32)pcID);
  m_listCSLIDEFace.InsertLast(pCSLIDEFace, 0);

  return TRUE;
}

CSLIDEFace *CSLIDEWorld::PeekFace(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEFace*)(m_hashCSLIDEFace.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertPoint(CSLIDEPoint *pCSLIDEPoint)
{
  CHAR *pcID;

  ASSERT( pCSLIDEPoint != NULL );

  pcID = pCSLIDEPoint->GetID();

  m_hashCSLIDEPoint.Insert(pCSLIDEPoint, (UINT_32)pcID);
  m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);

  return TRUE;
}

CSLIDEPoint *CSLIDEWorld::PeekPoint(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEPoint*)(m_hashCSLIDEPoint.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertSurface(CSLIDESurface *pCSLIDESurface)
{
  CHAR *pcID;

  ASSERT( pCSLIDESurface != NULL );

  pcID = pCSLIDESurface->GetID();

  m_hashCSLIDESurface.Insert(pCSLIDESurface, (UINT_32)pcID);
  m_listCSLIDESurface.InsertLast(pCSLIDESurface, 0);

  return TRUE;
}

CSLIDESurface *CSLIDEWorld::PeekSurface(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDESurface*)(m_hashCSLIDESurface.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertGlobalInitTclBlock(CHAR *pcTclBlock)
{
  ASSERT( pcTclBlock != NULL );

  m_listCSLIDEGlobalInitTclBlock.InsertLast(pcTclBlock, 0);

  return TRUE;
}

BOOL CSLIDEWorld::InsertGlobalUpdateTclBlock(CHAR *pcTclBlock)
{
  ASSERT( pcTclBlock != NULL );

  m_listCSLIDEGlobalUpdateTclBlock.InsertLast(pcTclBlock, 0);

  return TRUE;
}

BOOL CSLIDEWorld::CheckSemantics()
{
  BOOL b, bCorrect;
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDESurface *pCSLIDESurface;
  CHAR *pcTclBlock;
  INT_32 iCode;

  bCorrect = TRUE;

  // Create the toplevel tk widget before calling the Tcl Blocks
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->CreateTclToplevel();
    }

  fprintf(stderr, "BEGIN: Executing global tclinit blocks...\n");

  iter.Init(&m_listCSLIDEGlobalInitTclBlock);

  for ( pcTclBlock = (CHAR*)(iter.PeekFirst());
	pcTclBlock != NULL;
	pcTclBlock = (CHAR*)(iter.PeekNext()) )
    {
      iCode = Tcl_GlobalEval(m_pInterp, pcTclBlock);
      if ( iCode != TCL_OK )
	{
	  fprintf(CSLIDEPrimitive::GetErrorFile(), 
		  "Error: Tcl returned error code \"%s\" for the block:\n%s",
		  GetInterp()->result, pcTclBlock);
	  CSLIDEPrimitive::SetErrorCount(CSLIDEPrimitive::GetErrorCount()+1);
	  bCorrect = FALSE;
	}
    }

  fprintf(stderr, "\nEND: Executing global tclinit blocks\n");

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      b = pCSLIDESurface->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      b = pCSLIDEPoint->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      b = pCSLIDEFace->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      b = pCSLIDETransform->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      b = pCSLIDEInstance->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      b = pCSLIDENode->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  // Resolve Render Statements 
  if ( m_listCSLIDERender.GetLength() == 0 )
    {
      fprintf(CSLIDEPrimitive::GetErrorFile(), 
	      "Warning: No %s statement was defined so nothing will be rendered.\n", 
	      gcppcSLIDEReservedWords[sliderwtRender]);
      CSLIDEPrimitive::SetWarningCount(CSLIDEPrimitive::GetWarningCount()+1);
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      b = pCSLIDERender->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      b = pCWindow->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  return bCorrect;
}

BOOL CSLIDEWorld::Valid()
{
  return TRUE;
}

VOID CSLIDEWorld::Preprocess()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDESurface *pCSLIDESurface;

  UpdateTclBlocks();

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->Preprocess();
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->Preprocess();
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->Preprocess();
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      pCSLIDETransform->Preprocess();
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      pCSLIDEInstance->Preprocess();
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      pCSLIDENode->Preprocess();
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->Preprocess();
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->Preprocess();
    }

  PreprocessGraphs();

  CreateDynamicsList();

  // Preprocessing the Windows will actually create the windows and
  // bring them up on the screen.  This is why we do this last.
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->Preprocess();
    }

  // Get the dynamics going
  RestartDynamics();
  //PauseDynamics();
}

VOID CSLIDEWorld::PreprocessGraphs()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;

  // Run through the scene graphs of each of the render statements
  // to find which subtrees are static and to calculate all of the 
  // bounding boxes

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->PreprocessGraph();
    }
}

VOID CSLIDEWorld::CreateDynamicsList()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDESurface *pCSLIDESurface;

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      if ( pCSLIDESurface->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDESurface, 0);
	}
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      if ( pCSLIDEPoint->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEPoint, 0);
	}
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      if ( pCSLIDEFace->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEFace, 0);
	}
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      if ( pCSLIDETransform->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDETransform, 0);
	}
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      if ( pCSLIDEInstance->GetDynamic() || 
	   pCSLIDEInstance->GetTransformDynamic() || 
	   pCSLIDEInstance->GetBoundDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEInstance, 0);
	}
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      if ( pCSLIDENode->GetDynamic() ||
	   pCSLIDENode->GetBoundDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDENode, 0);
	}
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      if ( pCSLIDEViewport->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEViewport, 0);
	}
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      if ( pCSLIDERender->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDERender, 0);
	}
    }

  // Print out which objects were dynamic
/*
  printf("\n********* DYNAMIC OBJECTS **********\n");
  iter.Init(&m_listDynamics);
  CSLIDEPrimitive *pCSLIDEPrimitive;

  for ( pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekFirst());
	pCSLIDEPrimitive != NULL;
	pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekNext()) )
    {
      printf("%s: %ld: %ld: \"%s\"\n", pCSLIDEPrimitive->GetFileName(),
	     pCSLIDEPrimitive->GetFileLine(), 
	     pCSLIDEPrimitive->GetFileColumn(),
	     ( pCSLIDEPrimitive->GetID() == NULL ) ? ("No ID") :
	     (pCSLIDEPrimitive->GetID()));
    }
  printf("********* END: DYNAMIC OBJECTS **********\n");
*/
}

VOID CSLIDEWorld::Update()
{
  CListIterInt iter;
  CSLIDEPrimitive *pCSLIDEPrimitive;
  CWindow *pCWindow;

  UpdateTclBlocks();

  // Only call Update on the objects with some Dynamic aspect
  iter.Init(&m_listDynamics);

  for ( pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekFirst());
	pCSLIDEPrimitive != NULL;
	pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekNext()) )
    {
      pCSLIDEPrimitive->Update();
    }

  UpdateGraphs();

  // Update the windows after all other updates have happened
  // Updating a window == scheduling a Redisplay Event
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->Update();
    }
}

VOID CSLIDEWorld::UpdateTclBlocks()
{
  CListIterInt iter;
  CHAR *pcTclBlock;
  INT_32 iCode;

  iter.Init(&m_listCSLIDEGlobalUpdateTclBlock);

  for ( pcTclBlock = (CHAR*)(iter.PeekFirst());
	pcTclBlock != NULL;
	pcTclBlock = (CHAR*)(iter.PeekNext()) )
    {
      iCode = Tcl_GlobalEval(m_pInterp, pcTclBlock);
      if ( iCode != TCL_OK )
	{
	  fprintf(CSLIDEPrimitive::GetErrorFile(), 
		  "Error: Tcl returned error code \"%s\" for the block:\n%s",
		  GetInterp()->result, pcTclBlock);
	  exit(1);
	}
    }
}

VOID CSLIDEWorld::UpdateGraphs()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;

  // Run through the scene graphs of the render statements
  // and update bounding boxes
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->UpdateGraph();
    }
}

VOID CSLIDEWorld::RestartDynamics()
{
  // Remove the Timer Event from the Event Queue in case 
  // dynamics are currently running
  UnscheduleDynamics();

  // Initialize the system time to zero
  m_tvStart.RecordTime();
  m_dTimeBase = 0.0;
  m_uiFrame = 0;

  // Update the data structures at time zero, this will also
  // restart the timer events if we are not paused
  HandleDynamics();
}  

VOID CSLIDEWorld::PauseDynamics()
{
  m_bDynamicsPaused = !m_bDynamicsPaused;
  if ( m_bDynamicsPaused )
    {
      TimeValue tvEnd, tvDelta;

      // Pause the Dynamics
      UnscheduleDynamics();

      // Update the Base amount of time
      tvEnd.RecordTime();
      tvDelta = tvEnd - m_tvStart;
      m_dTimeBase += tvDelta.GetSeconds();
    }
  else
    {
      // Unpause the Dynamics
      
      // Record the new start time
      m_tvStart.RecordTime();

      // Schedule an Update for as soon as possible
      ScheduleDynamics(0);
    }
}

VOID CSLIDEWorld::ScheduleDynamics(INT_32 iSleepTime)
{
  //ASSERT( m_ttkWakeUpIDDynamics == 0 );
  if ( m_ttkWakeUpIDDynamics == 0 )
    {
      m_ttkWakeUpIDDynamics = Tk_CreateTimerHandler((int)iSleepTime, 
			        CSLIDEWorld::CBHandleDynamics, 
			        (ClientData)this);
    }
}

VOID CSLIDEWorld::UnscheduleDynamics()
{
  if ( m_ttkWakeUpIDDynamics != 0 )
    {
      Tk_DeleteTimerHandler(m_ttkWakeUpIDDynamics);
    }
  m_ttkWakeUpIDDynamics = 0;
}

VOID CSLIDEWorld::HandleDynamics()
{
  TimeValue tvStart, tvEnd, tvDelta;
  INT_32 iSleepTime;

  // The Timer Event has happened so the token is no longer valid
  m_ttkWakeUpIDDynamics = 0;

  tvStart.RecordTime();
  tvDelta = tvStart - m_tvStart;
  m_dUpdateTime = m_dTimeBase + tvDelta.GetSeconds();

  if ( !m_bDynamicsPaused )
    {
      // Update the value of the TCL SLF_TIME variable
      sprintf(gpcBuffer, "%f", m_dUpdateTime);
      Tcl_SetVar(m_pInterp, "SLF_TIME", gpcBuffer, TCL_GLOBAL_ONLY);
      m_uiFrame++;
      sprintf(gpcBuffer, "%ld", m_uiFrame);
      Tcl_SetVar(m_pInterp, "SLF_FRAME", gpcBuffer, TCL_GLOBAL_ONLY);
    }

  Update();

  tvEnd.RecordTime();
  tvDelta = tvEnd - tvStart;
  m_dUpdateTime = tvDelta.GetSeconds();

  iSleepTime = (INT_32)((m_dSecondsPerFrameRequested - m_dUpdateTime) * 1000.0);
  if ( iSleepTime < 0 )
    {
      iSleepTime = 0;
    }

  if ( !m_bDynamicsPaused )
    {
      //printf("Dynamics going to sleep for %ld ms.\n", iSleepTime);
      ScheduleDynamics(iSleepTime);
    }
}

VOID CSLIDEWorld::CBHandleDynamics(ClientData pvClientData)
{
  CSLIDEWorld *pCSLIDEWorld;

  ASSERT( pvClientData != NULL );

  pCSLIDEWorld = (CSLIDEWorld*)pvClientData;

  pCSLIDEWorld->HandleDynamics();
}

VOID CSLIDEWorld::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEFace *pCSLIDEFace;
  CSLIDENode *pCSLIDENode;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDERender *pCSLIDERender;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDESurface);

  // Skip the Default Surface
  pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
  ASSERT( pCSLIDESurface == CSLIDESurface::m_pCSLIDESurfaceDefault );
  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      pCSLIDENode->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->WriteSLIDE(fpOutputFile);
    }
}

UINT_32 CSLIDEWorld::GetErrorCount()
{
  return CSLIDEPrimitive::GetErrorCount();
}

UINT_32 CSLIDEWorld::GetWarningCount()
{
  return CSLIDEPrimitive::GetWarningCount();
}




